from rest_framework import permissions


class PlaybookPermissions(permissions.DjangoModelPermissions):
    """
    Object-level permission to only allow user in a group to view it.
    """

    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }

    def get_required_object_permissions(self, method, model_cls):
        kwargs = {
            'app_label': model_cls._meta.app_label,
            'model_name': model_cls._meta.model_name
        }
        return [perm % kwargs for perm in self.perms_map[method]]

    def has_permission(self, request, view):
        # Workaround to ensure DjangoModelPermissions are not applied
        # to the root view when using DefaultRouter.
        if getattr(view, '_ignore_model_permissions', False):
            return True

        try:
            queryset = view.get_queryset()
        except AttributeError:
            queryset = getattr(view, 'queryset', None)

        assert queryset is not None, (
            'Cannot apply DjangoModelPermissions on a view that '
            'does not have `.queryset` property or overrides the '
            '`.get_queryset()` method.')

        perms = self.get_required_permissions(request.method, queryset.model)

        # Custom permission for custom route
        if view.action == 'run':
            return request.user.has_perm('playbook.can_run_playbook')

        return (
            request.user and
            (request.user.is_authenticated() or not self.authenticated_users_only) and
            request.user.has_perms(perms)
        )



from .base import StorageBase
from django.conf import settings
from boto.s3.key import Key
import boto


class S3Plugin(StorageBase):
    def __init__(self, task_id):
        super(S3Plugin, self).__init__(task_id)
        self.s3_connection = boto.connect_s3(aws_access_key_id=settings.S3_ACCESS_KEY,
                                             aws_secret_access_key=settings.S3_SECRET_KEY)
        self.s3_filename = 'cyclosible-{task_id}.log'.format(task_id=task_id)
        self.bucket = self.s3_connection.get_bucket(settings.S3_BUCKET)

    def write_log(self):
        k = Key(self.bucket)
        k.key = self.s3_filename
        self.log_url = k.generate_url(expires_in=600, query_auth=False)
        if k.set_contents_from_file(self.tmpfile):
            return True
        else:
            return False



import abc
import six
import logging
logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class VaultBase(object):
    """Base class for Cyclosible vault plugin
    """

    def __init__(self):
        self.vault_password = None

    @abc.abstractmethod
    def get_password(self):
        """Get the ansible vault password.
        :returns: String
        """



from .base import VaultBase
from django.conf import settings


class PasswordPlugin(VaultBase):
    def get_password(self):
        self.vault_password = settings.VAULT_PASSWORD
        return self.vault_password



from django.contrib.auth.models import User
from celery.utils.log import get_task_logger
from django.conf import settings
from django.utils import timezone
from stevedore import enabled, driver
from ..Cyclosible.celery import app
from .models import Playbook
from .plugins.storage.base import check_plugin_enabled
from .callbacks_ansiblev1 import PlaybookCallbacks, PlaybookRunnerCallbacks, display
from .models import PlaybookRunHistory
from ansible import errors
from ansible import callbacks
from ansible import utils
from ansible.inventory import Inventory
from ansible.color import ANSIBLE_COLOR, stringc
import json
import ansible.playbook
import ansible.utils.template

logger = get_task_logger(__name__)


def colorize(lead, num, color):
    """ Print 'lead' = 'num' in 'color' """
    if num != 0 and ANSIBLE_COLOR and color is not None:
        return "%s%s%-15s" % (stringc(lead, color), stringc("=", color), stringc(str(num), color))
    else:
        return "%s=%-4s" % (lead, str(num))


def hostcolor(host, stats, color=True):
    if ANSIBLE_COLOR and color:
        if stats['failures'] != 0 or stats['unreachable'] != 0:
            return "%-37s" % stringc(host, 'red')
        elif stats['changed'] != 0:
            return "%-37s" % stringc(host, 'yellow')
        else:
            return "%-37s" % stringc(host, 'green')
    return "%-26s" % host


@app.task(bind=True, name="Run a playbook")
def run_playbook(self, playbook_name, user_name, only_tags=None, skip_tags=None, extra_vars=None, subset=None):
    """ This function will launch a playbook. To handle logging, it will
    use stevedore which will load all extensions registered under the
    entrypoint cyclosible.plugins.storage. For example, it will let Cyclosible
    save his log on a file, on S3, or something else.
    :param playbook_name:
    :param user_name:
    :param only_tags:
    :param skip_tags:
    :param extra_vars:
    :return:
    """
    history = PlaybookRunHistory.objects.create(
        playbook=Playbook.objects.get(name=playbook_name),
        date_launched=timezone.now(),
        status='RUNNING',
        task_id=self.request.id,
        launched_by=User.objects.get(username=user_name)
    )

    vault_password = None
    if settings.VAULT_ENABLED:
        try:
            self.mgr_vault = driver.DriverManager(
                namespace='cyclosible.plugins.vault',
                name=settings.VAULT_ENABLED,
                invoke_on_load=True,
            )
            vault_password = self.mgr_vault.driver.get_password()
        except RuntimeError as e:
            logger.error(e)

        logger.debug('LOADED VAULT: {plugins} | Status: {status}'.format(
            plugins=settings.VAULT_ENABLED,
            status='OK' if vault_password else 'KO'
        ))

    inventory = Inventory(vault_password=vault_password)

    # Here, we override the default ansible callbacks to pass our customs parameters
    stats = callbacks.AggregateStats()
    playbook_cb = PlaybookCallbacks(
        verbose=utils.VERBOSITY,
        task_id=self.request.id
    )
    runner_cb = PlaybookRunnerCallbacks(
        stats=stats,
        verbose=utils.VERBOSITY,
        task_id=self.request.id
    )

    pb = ansible.playbook.PlayBook(
        playbook=''.join([settings.PLAYBOOK_PATH, playbook_name, '.yml']),
        callbacks=playbook_cb,
        runner_callbacks=runner_cb,
        inventory=inventory,
        stats=stats,
        extra_vars=extra_vars,
        only_tags=only_tags,
        skip_tags=skip_tags,
        vault_password=vault_password,
        subset=subset
    )

    try:
        pb.run()
        hosts = sorted(pb.stats.processed.keys())
        logger.info(hosts)
        playbook_cb.on_stats(pb.stats)
        history.status = 'SUCCESS'
        for h in hosts:
            t = pb.stats.summarize(h)

            display("%s : %s %s %s %s" % (
                hostcolor(h, t),
                colorize('ok', t['ok'], 'green'),
                colorize('changed', t['changed'], 'yellow'),
                colorize('unreachable', t['unreachable'], 'red'),
                colorize('failed', t['failures'], 'red')),
                task_id=self.request.id
            )
    except errors.AnsibleError as e:
        history.status = 'FAILED'
        display(u"ERROR: %s" % e, task_id=self.request.id)
        logger.error(u"ERROR: %s" % e)

    try:
        self.mgr_storage = enabled.EnabledExtensionManager(
            namespace='cyclosible.plugins.storage',
            check_func=check_plugin_enabled,
            invoke_on_load=True,
            invoke_kwds={'task_id': self.request.id},
            verify_requirements=True
        )

        logger.debug('LOADED STORAGE: {plugins}'.format(plugins=', '.join(self.mgr_storage.names())))
    except RuntimeError as e:
        logger.error(e)

    try:
        list_urls = []
        self.mgr_storage.map(lambda ext: (ext.name, ext.obj.write_log()))
        urls = self.mgr_storage.map(lambda ext: (ext.name, ext.obj.get_url_log()))
        for url in urls:
            try:
                if url[1]:
                    list_urls.append({url[0]: url[1]})
            except IndexError:
                logger.debug('Index does not exist in the url returned')
        history.log_url = json.dumps(list_urls)
    except RuntimeError:
        logger.debug('No plugins available')
    history.date_finished = timezone.now()
    history.save()
    if history.status == 'FAILED':
        return 1



from .models import Playbook, PlaybookRunHistory
from django.contrib.auth.models import Group
from rest_framework import serializers
from ..Cyclosible.serializers import GroupSerializerNoValidator
from django.core.exceptions import ObjectDoesNotExist


class PlaybookRunHistorySerializer(serializers.HyperlinkedModelSerializer):
    playbook = serializers.SlugRelatedField(
        read_only=True,
        slug_field='name'
    )

    launched_by = serializers.SlugRelatedField(
        read_only=True,
        slug_field='username'
    )

    class Meta:
        model = PlaybookRunHistory
        fields = ('url', 'playbook', 'date_launched', 'date_finished', 'status', 'task_id', 'launched_by', 'log_url')


class FilteredHistorySerializer(serializers.ListSerializer):

    def to_representation(self, data):
        data = data.order_by('-date_launched')[:10]
        return super(FilteredHistorySerializer, self).to_representation(data)


class PlaybookRunHistoryLastSerializer(PlaybookRunHistorySerializer):
    class Meta:
        model = PlaybookRunHistory
        list_serializer_class = FilteredHistorySerializer


class PlaybookSerializer(serializers.HyperlinkedModelSerializer):
    group = GroupSerializerNoValidator(many=False, required=True)
    history = PlaybookRunHistoryLastSerializer(many=True, read_only=True)

    class Meta:
        model = Playbook
        fields = ('url', 'name', 'only_tags', 'skip_tags', 'extra_vars', 'subset', 'group', 'history')
        extra_kwargs = {
            'url': {'lookup_field': 'name'}
        }

    def create(self, validated_data):
        group_data = validated_data.pop('group')
        try:
            group = Group.objects.get(name=group_data.get('name'))
        except Group.DoesNotExist:
            group = Group.objects.create(**group_data)
        playbook = Playbook.objects.create(group=group, **validated_data)
        return playbook

    def update(self, instance, validated_data):
        instance.name = validated_data.get('name', instance.name)
        instance.only_tags = validated_data.get('only_tags', instance.only_tags)
        instance.skip_tags = validated_data.get('skip_tags', instance.skip_tags)
        instance.extra_vars = validated_data.get('extra_vars', instance.skip_tags)
        instance.subset = validated_data.get('subset', instance.skip_tags)
        group_data = validated_data.get('group', {'name': instance.group.name})
        try:
            group = Group.objects.get(name=group_data.get('name'))
            instance.group = group
        except Group.DoesNotExist:
            raise ObjectDoesNotExist
        return instance


class RunPlaybookSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Playbook
        fields = ('url', 'only_tags', 'skip_tags', 'extra_vars', 'subset')
        extra_kwargs = {
            'url': {'lookup_field': 'name'}
        }



"""
WSGI config for Cyclosible project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/howto/deployment/wsgi/
"""

import os
from django.core.wsgi import get_wsgi_application
from ws4redis.uwsgi_runserver import uWSGIWebsocketServer
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "Cyclosible.settings")

_django_app = get_wsgi_application()
_websocket_app = uWSGIWebsocketServer()


def application(environ, start_response):
    from django.conf import settings
    if environ.get('PATH_INFO').startswith(settings.WEBSOCKET_URL):
        return _websocket_app(environ, start_response)
    return _django_app(environ, start_response)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('playbook', '0004_playbookrunhistory_log_url'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='playbook',
            options={'permissions': (('view_playbook', 'Can view the playbook'), ('can_override_skip_tags', 'Can override skip_tags'), ('can_override_only_tags', 'Can override only_tags'), ('can_override_extra_vars', 'Can override extra_vars'), ('can_run_playbook', 'Can run the playbook'))},
        ),
        migrations.AddField(
            model_name='playbook',
            name='extra_vars',
            field=models.CharField(default='', max_length=1024, blank=True),
        ),
        migrations.AlterField(
            model_name='playbook',
            name='only_tags',
            field=models.CharField(default='', max_length=1024, blank=True),
        ),
        migrations.AlterField(
            model_name='playbook',
            name='skip_tags',
            field=models.CharField(default='', max_length=1024, blank=True),
        ),
        migrations.AlterField(
            model_name='playbookrunhistory',
            name='log_url',
            field=models.CharField(default='', max_length=1024, blank=True),
        ),
        migrations.AlterField(
            model_name='playbookrunhistory',
            name='status',
            field=models.CharField(default='RUNNING', max_length=1024),
        ),
        migrations.AlterField(
            model_name='playbookrunhistory',
            name='task_id',
            field=models.CharField(default='', max_length=1024, blank=True),
        ),
    ]



from rest_framework.routers import DefaultRouter


class ContainerRouter(DefaultRouter):
    def register_router(self, router):
        self.registry.extend(router.registry)

main_router = ContainerRouter()



__title__ = 'Cyclosible'
__version__ = '0.3'
__author__ = 'Cycloid'
__license__ = 'Apache'
__copyright__ = 'Copyright 2015'

# Version synonym
VERSION = __version__

# Header encoding (see RFC5987)
HTTP_HEADER_ENCODING = 'iso-8859-1'

# Default datetime input and output formats
ISO_8601 = 'iso-8601'



from django.contrib.auth.models import User, Group
from rest_framework import serializers
import logging
logger = logging.getLogger(__name__)


class GroupSerializerNoValidator(serializers.HyperlinkedModelSerializer):
    name = serializers.SlugField(validators=[])

    class Meta:
        model = Group
        lookup_field = 'name'
        fields = ('url', 'name')
        extra_kwargs = {
            'url': {'lookup_field': 'name'}
        }


class GroupSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = Group
        lookup_field = 'name'
        fields = ('url', 'name')
        extra_kwargs = {
            'url': {'lookup_field': 'name'}
        }


class UserSerializer(serializers.HyperlinkedModelSerializer):
    groups = GroupSerializerNoValidator(many=True, required=False)

    class Meta:
        model = User
        lookup_field = 'username'
        fields = ('url', 'first_name', 'last_name', 'username', 'email', 'groups')
        extra_kwargs = {
            'url': {'lookup_field': 'username'}
        }

    def create(self, validated_data):
        groups_data = validated_data.pop('groups')
        user = User.objects.create(**validated_data)
        for group_data in groups_data:
            try:
                group = Group.objects.get(name=group_data.get('name'))
                group.user_set.add(user)
            except Group.DoesNotExist:
                group = Group.objects.create(**group_data)
                group.user_set.add(user)
        return user

    def update(self, instance, validated_data):
        instance.first_name = validated_data.get('first_name', instance.first_name)
        instance.last_name = validated_data.get('last_name', instance.last_name)
        instance.username = validated_data.get('username', instance.username)
        instance.email = validated_data.get('email', instance.email)
        instance.groups.clear()
        groups_data = validated_data.get('groups')
        for group_data in groups_data:
            try:
                group = Group.objects.get(name=group_data.get('name'))
                group.user_set.add(instance)
            except Group.DoesNotExist:
                group = Group.objects.create(**group_data)
                group.user_set.add(instance)
        return instance



"""
WSGI config for Cyclosible project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/howto/deployment/wsgi/
"""

from ws4redis.uwsgi_runserver import uWSGIWebsocketServer
import os
import gevent.socket
import redis.connection


redis.connection.socket = gevent.socket
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "Cyclosible.settings")
application = uWSGIWebsocketServer()



from django.conf import settings
from redis import Redis, ConnectionPool
import abc
import six
import logging
import tempfile
logger = logging.getLogger(__name__)


def check_plugin_enabled(plugin):
    if plugin.name in settings.STORAGE_ENABLED:
        return True
    else:
        return False


@six.add_metaclass(abc.ABCMeta)
class StorageBase(object):
    """Base class for Cyclosible plugin
    """

    def __init__(self, task_id):
        self.redis_connection = Redis(connection_pool=ConnectionPool(**settings.WS4REDIS_CONNECTION))
        self.tmpfile = tempfile.NamedTemporaryFile(mode='a+')
        for log in self.redis_connection.lrange(':'.join(['tasks', task_id]), 0, -1):
            self.tmpfile.write(log)
        # Rewind the log file
        self.tmpfile.seek(0)
        self.log_url = None

    @abc.abstractmethod
    def write_log(self):
        """Write the log somewhere.
        :returns: Boolean. True if success, False if failure
        """

    def get_url_log(self):
        """Get the url of the log.
        :returns: Url of the log. Can be empty
        """
        return self.log_url



from .base import VaultBase
from django.conf import settings


class FilePlugin(VaultBase):
    def get_password(self):
        with open(settings.VAULT_FILE) as vaultfile:
            self.vault_password = vaultfile.read()
            return self.vault_password



from rest_framework import viewsets
from .models import Playbook, PlaybookRunHistory
from .serializers import (PlaybookSerializer, PlaybookRunHistorySerializer)
from rest_framework.decorators import detail_route
from rest_framework.response import Response
from rest_framework import status
from rest_framework import filters
from rest_framework.permissions import IsAdminUser
from .permissions import PlaybookPermissions
from .tasks import run_playbook as task_run_playbook
from .serializers import RunPlaybookSerializer
import ansible.constants as C


class PlaybookViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows playbooks to be viewed or edited.
    """
    queryset = Playbook.objects.all()
    serializer_class = PlaybookSerializer
    lookup_field = 'name'
    permission_classes = (PlaybookPermissions,)
    filter_backends = (filters.DjangoObjectPermissionsFilter,)

    @detail_route(methods=['POST'], serializer_class=RunPlaybookSerializer)
    def run(self, request, name):
        """
        This function launch an ansible playbook in celery. The user must have the permission can_run_playbook
        :param request:
        :param name:
        :return:

        ---
        request_serializer: RunPlaybookSerializer
        response_serializer: RunPlaybookSerializer
        """

        if request.method == 'POST':
            playbook = self.get_object()
            if request.user.has_perm('playbook.can_run_playbook', playbook):
                serializer = RunPlaybookSerializer(data=request.data)
                if serializer.is_valid():
                    only_tags = None
                    skip_tags = None
                    extra_vars = None

                    # Test permissions
                    if 'only_tags' in serializer.validated_data:
                        if request.user.has_perm('playbook.can_override_only_tags', playbook):
                            only_tags = serializer.validated_data.get('only_tags').split(',')
                        else:
                            return Response(
                                {'status': 'You have not the permission to override tags on this playbook'},
                                status=status.HTTP_403_FORBIDDEN
                            )
                    else:
                        if playbook.only_tags:
                            only_tags = playbook.only_tags.split(',')

                    if 'skip_tags' in serializer.validated_data:
                        if request.user.has_perm('playbook.can_override_skip_tags', playbook):
                            skip_tags = serializer.validated_data.get('skip_tags').split(',')
                        else:
                            return Response(
                                {'status': 'You have not the permission to override tags on this playbook'},
                                status=status.HTTP_403_FORBIDDEN
                            )
                    else:
                        if playbook.skip_tags:
                            skip_tags = playbook.skip_tags.split(',')

                    if 'subset' in serializer.validated_data:
                        if request.user.has_perm('playbook.can_override_subset', playbook):
                            subset = serializer.validated_data.get('subset')
                        else:
                            return Response(
                                {'status': 'You have not the permission to override subset on this playbook'},
                                status=status.HTTP_403_FORBIDDEN
                            )
                    else:
                        if playbook.subset:
                            subset = playbook.subset
                        else:
                            subset = C.DEFAULT_SUBSET

                    if 'extra_vars' in serializer.validated_data:
                        if request.user.has_perm('playbook.can_override_extra_vars', playbook):
                            extra_vars = {}
                            for var in serializer.validated_data.get('extra_vars').split(','):
                                extra = var.split('=')
                                extra_vars[extra[0]] = extra[1]
                        else:
                            return Response(
                                {'status': 'You have not the permission to override extra_vars on this playbook'},
                                status=status.HTTP_403_FORBIDDEN
                            )
                    else:
                        if playbook.extra_vars:
                            extra_vars = {}
                            for var in playbook.extra_vars.split(','):
                                extra = var.split('=')
                                extra_vars[extra[0]] = extra[1]

                    # Launch the playbook
                    task = task_run_playbook.delay(
                        playbook_name=playbook.name,
                        user_name=request.user.username,
                        only_tags=only_tags,
                        skip_tags=skip_tags,
                        extra_vars=extra_vars,
                        subset=subset
                    )

                    # Return a response to the api request
                    return Response(
                        {
                            'status': 'Playbook has been launched',
                            'task_id': task.id,
                            'websocket_url': 'ws://{server_name}:{server_port}/ws/{task_id}?subscribe-broadcast'.format(
                                server_name=request.META.get('SERVER_NAME'),
                                server_port=request.META.get('SERVER_PORT'),
                                task_id=task.id
                            ),
                        },
                        status=status.HTTP_201_CREATED
                    )
                else:
                    return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
            else:
                return Response({'status': 'You have not the permission to run this playbook'},
                                status=status.HTTP_403_FORBIDDEN)


class PlaybookRunHistoryViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows playbooks run history to be viewed only.
    """
    queryset = PlaybookRunHistory.objects.all()
    serializer_class = PlaybookRunHistorySerializer
    permission_classes = (IsAdminUser,)



from django.db import models
from django.contrib.auth.models import Group, User

from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver
from rest_framework.authtoken.models import Token


class Playbook(models.Model):
    class Meta:
        app_label = 'playbook'
        permissions = (
            ("view_playbook", "Can view the playbook"),
            ("can_override_skip_tags", 'Can override skip_tags'),
            ("can_override_only_tags", 'Can override only_tags'),
            ("can_override_extra_vars", 'Can override extra_vars'),
            ("can_override_subset", 'Can override subset'),
            ("can_run_playbook", 'Can run the playbook')
        )

    name = models.CharField(max_length=100, unique=True, db_index=True)
    only_tags = models.CharField(max_length=1024, default='', blank=True)
    skip_tags = models.CharField(max_length=1024, default='', blank=True)
    extra_vars = models.CharField(max_length=1024, default='', blank=True)
    subset = models.CharField(max_length=1024, default='', blank=True)
    group = models.ForeignKey(Group, null=True)

    def __unicode__(self):
        return self.name


class PlaybookRunHistory(models.Model):
    class Meta:
        app_label = 'playbook'

    STATUS = (
        ('SUCCESS', 'Success'),
        ('RUNNING', 'Running'),
        ('FAILURE', 'Failure'),
    )

    playbook = models.ForeignKey(Playbook, db_index=True, related_name='history')
    date_launched = models.DateTimeField(blank=True)
    date_finished = models.DateTimeField(blank=True, null=True)
    status = models.CharField(max_length=1024, default='RUNNING')
    task_id = models.CharField(max_length=1024, default='', blank=True)
    log_url = models.CharField(max_length=1024, default='', blank=True)
    launched_by = models.ForeignKey(User)


@receiver(post_save, sender=settings.AUTH_USER_MODEL)
def create_auth_token(sender, instance=None, created=False, **kwargs):
    if created:
        Token.objects.create(user=instance)



from .base import VaultBase
from django.conf import settings
import hvac


class HashicorpPlugin(VaultBase):
    def get_password(self):
        client = hvac.Client(
            url='{scheme}://{host}:{port}'.format(
                scheme=settings.VAULT_HASHICORP.get('scheme'),
                host=settings.VAULT_HASHICORP.get('host'),
                port=settings.VAULT_HASHICORP.get('port'),
            ),
            token=settings.VAULT_HASHICORP.get('token')
        )
        self.vault_password = client.read(
            settings.VAULT_HASHICORP.get('secret_path')
        ).get('data').get(settings.VAULT_HASHICORP.get('secret_field'))
        return self.vault_password



from django.contrib.auth.models import User, Group
from rest_framework import viewsets
from .serializers import (UserSerializer, GroupSerializer)
from django.http import HttpResponseForbidden


class UserViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows users to be viewed or edited.
    """
    queryset = User.objects.all()
    serializer_class = UserSerializer
    lookup_field = 'username'


class GroupViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows groups to be viewed or edited.
    """
    queryset = Group.objects.all()
    serializer_class = GroupSerializer
    lookup_field = 'name'


def permission_denied_handler(request):
    return HttpResponseForbidden('You have no permissions!')



from ws4redis.subscriber import RedisSubscriber


class WebSocketSubscriber(RedisSubscriber):
    def __init__(self, connection):
        self._subscription = None
        super(WebSocketSubscriber, self).__init__(connection)

    def send_persited_messages(self, websocket):
        """
        This method is called immediately after a websocket is openend by the client, so that
        persisted messages can be sent back to the client upon connection.
        """
        for channel in self._subscription.channels:
            task_id = channel.split(':')[2]
            task_channel = ':'.join(['tasks', task_id])
            logs = self._connection.lrange(task_channel, 0, -1)[:-1]
            for log in logs:
                websocket.send(log)

            message = self._connection.get(channel)
            if message:
                websocket.send(message)



#!/usr/bin/env python
import os
import sys


def main():
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "cyclosible.Cyclosible.settings")
    from django.core.management import execute_from_command_line
    execute_from_command_line(sys.argv)

if __name__ == "__main__":
    main()



from ansible import utils
from ansible import constants
from ansible.module_utils import basic
from ansible.utils.unicode import to_unicode, to_bytes
from ansible.callbacks import (DefaultRunnerCallbacks, call_callback_module, banner, log_flock, log_unflock)
from ansible.color import stringc
from ws4redis.publisher import RedisPublisher
from ws4redis.redis_store import RedisMessage
from django.conf import settings
from redis import Redis, ConnectionPool
import logging
import sys
import getpass
import fnmatch
import locale

redis_connection = Redis(connection_pool=ConnectionPool(**settings.WS4REDIS_CONNECTION))
logger = logging.getLogger(__name__)


def display(msg, task_id, color=None, stderr=False, screen_only=False, log_only=False, runner=None):
    # prevent a very rare case of interlaced multiprocess I/O
    log_flock(runner)
    while msg.startswith("\n"):
        msg = msg.replace("\n", "")
    msg2 = msg + '\n'
    if color:
        msg2 = stringc(msg, color) + '\n'

    logger.debug('TASK_ID: {task_id} | MSG: {message}'.format(task_id=task_id, message=msg2))

    # Pusblish the message on websocket
    redis_publisher = RedisPublisher(facility=task_id, broadcast=True)
    redis_message = RedisMessage(msg2)
    redis_publisher.publish_message(message=redis_message)

    # Store the message into a redis list to let the user get all
    # history of the logs
    redis_connection.rpush('tasks:' + task_id, msg2)
    # reset the expire of the list
    if settings.WS4REDIS_EXPIRE:
        if settings.WS4REDIS_EXPIRE > 0:
            redis_connection.expire('tasks:' + task_id,
                                    settings.WS4REDIS_EXPIRE)
    log_unflock(runner)


class PlaybookRunnerCallbacks(DefaultRunnerCallbacks):
    """ callbacks used for Runner() from /usr/bin/ansible-playbook """

    def __init__(self, stats, task_id, verbose=None):

        if verbose is None:
            verbose = utils.VERBOSITY

        self.task_id = task_id
        self.verbose = verbose
        self.stats = stats
        self._async_notified = {}
        super(DefaultRunnerCallbacks, self).__init__()

    def on_unreachable(self, host, results):
        if self.runner.delegate_to:
            host = '%s -> %s' % (host, self.runner.delegate_to)

        item = None
        if type(results) == dict:
            item = results.get('item', None)
            if isinstance(item, unicode):
                item = utils.unicode.to_bytes(item)
            results = basic.json_dict_unicode_to_bytes(results)
        else:
            results = utils.unicode.to_bytes(results)
        host = utils.unicode.to_bytes(host)
        if item:
            msg = "fatal: [%s] => (item=%s) => %s" % (host, item, results)
        else:
            msg = "fatal: [%s] => %s" % (host, results)
        display(msg, color='red', runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_unreachable(host, results)

    def on_failed(self, host, results, ignore_errors=False):
        if self.runner.delegate_to:
            host = '%s -> %s' % (host, self.runner.delegate_to)

        results2 = results.copy()
        results2.pop('invocation', None)

        item = results2.get('item', None)
        parsed = results2.get('parsed', True)
        module_msg = ''
        if not parsed:
            module_msg = results2.pop('msg', None)
        stderr = results2.pop('stderr', None)
        stdout = results2.pop('stdout', None)
        returned_msg = results2.pop('msg', None)

        if item:
            msg = "failed: [%s] => (item=%s) => %s" % (host, item, utils.jsonify(results2))
        else:
            msg = "failed: [%s] => %s" % (host, utils.jsonify(results2))
        display(msg, color='red', runner=self.runner, task_id=self.task_id)

        if stderr:
            display("stderr: %s" % stderr, color='red', runner=self.runner, task_id=self.task_id)
        if stdout:
            display("stdout: %s" % stdout, color='red', runner=self.runner, task_id=self.task_id)
        if returned_msg:
            display("msg: %s" % returned_msg, color='red', runner=self.runner, task_id=self.task_id)
        if not parsed and module_msg:
            display(module_msg, color='red', runner=self.runner, task_id=self.task_id)
        if ignore_errors:
            display("...ignoring", color='cyan', runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_failed(host, results, ignore_errors=ignore_errors)

    def on_ok(self, host, host_result):
        if self.runner.delegate_to:
            host = '%s -> %s' % (host, self.runner.delegate_to)

        item = host_result.get('item', None)

        host_result2 = host_result.copy()
        host_result2.pop('invocation', None)
        verbose_always = host_result2.pop('verbose_always', False)
        changed = host_result.get('changed', False)
        ok_or_changed = 'ok'
        if changed:
            ok_or_changed = 'changed'

        # show verbose output for non-setup module results if --verbose is used
        msg = ''
        if (not self.verbose or host_result2.get("verbose_override", None) is not
                None) and not verbose_always:
            if item:
                msg = "%s: [%s] => (item=%s)" % (ok_or_changed, host, item)
            else:
                if 'ansible_job_id' not in host_result or 'finished' in host_result:
                    msg = "%s: [%s]" % (ok_or_changed, host)
        else:
            # verbose ...
            if item:
                msg = "%s: [%s] => (item=%s) => %s" % (ok_or_changed, host, item,
                                                       utils.jsonify(host_result2, format=verbose_always))
            else:
                if 'ansible_job_id' not in host_result or 'finished' in host_result2:
                    msg = "%s: [%s] => %s" % (ok_or_changed, host, utils.jsonify(host_result2, format=verbose_always))

        if msg != '':
            if not changed:
                display(msg, color='green', runner=self.runner, task_id=self.task_id)
            else:
                display(msg, color='yellow', runner=self.runner, task_id=self.task_id)
        if constants.COMMAND_WARNINGS and 'warnings' in host_result2 and host_result2['warnings']:
            for warning in host_result2['warnings']:
                display("warning: %s" % warning, color='purple', runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_ok(host, host_result)

    def on_skipped(self, host, item=None):
        if self.runner.delegate_to:
            host = '%s -> %s' % (host, self.runner.delegate_to)
        if constants.DISPLAY_SKIPPED_HOSTS:
            msg = ''
            if item:
                msg = "skipping: [%s] => (item=%s)" % (host, item)
            else:
                msg = "skipping: [%s]" % host
            display(msg, color='cyan', runner=self.runner, task_id=self.task_id)
            super(PlaybookRunnerCallbacks, self).on_skipped(host, item)

    def on_no_hosts(self):
        display("FATAL: no hosts matched or all hosts have already failed -- aborting\n", color='red',
                runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_no_hosts()

    def on_async_poll(self, host, res, jid, clock):
        if jid not in self._async_notified:
            self._async_notified[jid] = clock + 1
        if self._async_notified[jid] > clock:
            self._async_notified[jid] = clock
            msg = "<job %s> polling, %ss remaining" % (jid, clock)
            display(msg, color='cyan', runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_async_poll(host, res, jid, clock)

    def on_async_ok(self, host, res, jid):
        if jid:
            msg = "<job %s> finished on %s" % (jid, host)
            display(msg, color='cyan', runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_async_ok(host, res, jid)

    def on_async_failed(self, host, res, jid):
        msg = "<job %s> FAILED on %s" % (jid, host)
        display(msg, color='red', stderr=True, runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_async_failed(host, res, jid)

    def on_file_diff(self, host, diff):
        display(utils.get_diff(diff), runner=self.runner, task_id=self.task_id)
        super(PlaybookRunnerCallbacks, self).on_file_diff(host, diff)


class PlaybookCallbacks(object):
    """ playbook.py callbacks used by /usr/bin/ansible-playbook """

    def __init__(self, task_id, verbose=False):
        self.task_id = task_id
        self.verbose = verbose

    def on_start(self):
        call_callback_module('playbook_on_start')

    def on_notify(self, host, handler):
        call_callback_module('playbook_on_notify', host, handler)

    def on_no_hosts_matched(self):
        display("skipping: no hosts matched", color='cyan', task_id=self.task_id)
        call_callback_module('playbook_on_no_hosts_matched')

    def on_no_hosts_remaining(self):
        display("\nFATAL: all hosts have already failed -- aborting", color='red', task_id=self.task_id)
        call_callback_module('playbook_on_no_hosts_remaining')

    def on_task_start(self, name, is_conditional):
        name = utils.unicode.to_bytes(name)
        msg = "TASK: [%s]" % name
        if is_conditional:
            msg = "NOTIFIED: [%s]" % name

        if hasattr(self, 'start_at'):
            self.start_at = utils.unicode.to_bytes(self.start_at)
            if name == self.start_at or fnmatch.fnmatch(name, self.start_at):
                # we found out match, we can get rid of this now
                del self.start_at
            elif self.task.role_name:
                # handle tasks prefixed with rolenames
                actual_name = name.split('|', 1)[1].lstrip()
                if actual_name == self.start_at or fnmatch.fnmatch(actual_name, self.start_at):
                    del self.start_at

        if hasattr(self, 'start_at'):  # we still have start_at so skip the task
            self.skip_task = True
        elif hasattr(self, 'step') and self.step:
            if isinstance(name, str):
                name = utils.unicode.to_unicode(name)
            msg = u'Perform task: %s (y/n/c): ' % name
            if sys.stdout.encoding:
                msg = to_bytes(msg, sys.stdout.encoding)
            else:
                msg = to_bytes(msg)
            resp = raw_input(msg)
            if resp.lower() in ['y', 'yes']:
                self.skip_task = False
                display(banner(msg), task_id=self.task_id)
            elif resp.lower() in ['c', 'continue']:
                self.skip_task = False
                self.step = False
                display(banner(msg), task_id=self.task_id)
            else:
                self.skip_task = True
        else:
            self.skip_task = False
            display(banner(msg), task_id=self.task_id)

        call_callback_module('playbook_on_task_start', name, is_conditional)

    def on_vars_prompt(self, varname, private=True, prompt=None, encrypt=None, confirm=False, salt_size=None,
                       salt=None, default=None):

        if prompt and default is not None:
            msg = "%s [%s]: " % (prompt, default)
        elif prompt:
            msg = "%s: " % prompt
        else:
            msg = 'input for %s: ' % varname

        def do_prompt(prompt, private):
            if sys.stdout.encoding:
                msg = prompt.encode(sys.stdout.encoding)
            else:
                # when piping the output, or at other times when stdout
                # may not be the standard file descriptor, the stdout
                # encoding may not be set, so default to something sane
                msg = prompt.encode(locale.getpreferredencoding())
            if private:
                return getpass.getpass(msg)
            return raw_input(msg)

        if confirm:
            while True:
                result = do_prompt(msg, private)
                second = do_prompt("confirm " + msg, private)
                if result == second:
                    break
                display("***** VALUES ENTERED DO NOT MATCH ****", task_id=self.task_id)
        else:
            result = do_prompt(msg, private)

        # if result is false and default is not None
        if not result and default is not None:
            result = default

        if encrypt:
            result = utils.do_encrypt(result, encrypt, salt_size, salt)

        # handle utf-8 chars
        result = to_unicode(result, errors='strict')
        call_callback_module('playbook_on_vars_prompt', varname, private=private, prompt=prompt,
                             encrypt=encrypt, confirm=confirm, salt_size=salt_size, salt=None, default=default
                             )

        return result

    def on_setup(self):
        display(banner("GATHERING FACTS"), task_id=self.task_id)
        call_callback_module('playbook_on_setup')

    def on_import_for_host(self, host, imported_file):
        msg = "%s: importing %s" % (host, imported_file)
        display(msg, color='cyan', task_id=self.task_id)
        call_callback_module('playbook_on_import_for_host', host, imported_file)

    def on_not_import_for_host(self, host, missing_file):
        msg = "%s: not importing file: %s" % (host, missing_file)
        display(msg, color='cyan', task_id=self.task_id)
        call_callback_module('playbook_on_not_import_for_host', host, missing_file)

    def on_play_start(self, name):
        display(banner("PLAY [%s]" % name), task_id=self.task_id)
        call_callback_module('playbook_on_play_start', name)

    def on_stats(self, stats):
        call_callback_module('playbook_on_stats', stats)



"""Cyclosible URL Configuration
"""
from django.conf.urls import include, url
from django.contrib import admin
from rest_framework.authtoken import views as authtoken_views
from .views import (GroupViewSet, UserViewSet)
from .routers import main_router
from cyclosible.playbook import urls as playbook_urls
assert playbook_urls

main_router.register(r'groups', GroupViewSet)
main_router.register(r'users', UserViewSet)

urlpatterns = [
    url(r'^admin/', include(admin.site.urls)),
    url(r'^api/v1/', include(main_router.urls)),
    url(r'^api-auth/', include('rest_framework.urls', namespace='rest_framework')),
    url(r'^api-token-auth/', authtoken_views.obtain_auth_token),
    url(r'^docs/', include('rest_framework_swagger.urls')),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0006_require_contenttypes_0002'),
    ]

    operations = [
        migrations.CreateModel(
            name='Playbook',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=100, db_index=True)),
                ('only_tags', models.CharField(default=b'', max_length=1024, blank=True)),
                ('skip_tags', models.CharField(default=b'', max_length=1024, blank=True)),
                ('group', models.ForeignKey(to='auth.Group', null=True)),
            ],
            options={
                'permissions': (('view_playbook', 'Can view the playbook'), ('can_override_skip_tags', 'Can override skip_tags'), ('can_override_only_tags', 'Can override only_tags'), ('can_run_playbook', 'Can run the playbook')),
            },
        ),
        migrations.CreateModel(
            name='PlaybookRunHistory',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date_launched', models.DateTimeField(blank=True)),
                ('date_finished', models.DateTimeField(blank=True)),
                ('status', models.CharField(default=b'RUNNING', max_length=1024)),
                ('task_id', models.CharField(default=b'', max_length=1024, blank=True)),
                ('launched_by', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
                ('playbook', models.ForeignKey(to='playbook.Playbook')),
            ],
        ),
    ]



from django.contrib import admin
from .models import Playbook, PlaybookRunHistory
from guardian.admin import GuardedModelAdmin


@admin.register(Playbook)
class PlaybookAdmin(GuardedModelAdmin):
    queryset = Playbook.objects.all()
    list_display = ('name', 'group', 'only_tags', 'skip_tags')


@admin.register(PlaybookRunHistory)
class PlaybookRunHistoryAdmin(GuardedModelAdmin):
    queryset = PlaybookRunHistory.objects.all()
    list_display = ('playbook', 'date_launched', 'date_finished', 'status', 'task_id', 'launched_by')



"""
Django settings for Cyclosible project.

Generated by 'django-admin startproject' using Django 1.8.5.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

STATIC_ROOT = (os.path.join(BASE_DIR, "static"))
# Application definition

INSTALLED_APPS = (
    'djcelery',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'guardian',
    'cyclosible.playbook',
    'rest_framework',
    'rest_framework_swagger',
    'rest_framework.authtoken',
    'ws4redis',
)

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',  # default
    'guardian.backends.ObjectPermissionBackend',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
)

ROOT_URLCONF = 'cyclosible.Cyclosible.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

SWAGGER_SETTINGS = {
    'is_authenticated': True,
    'is_superuser': False,
    'permission_denied_handler': 'cyclosible.Cyclosible.views.permission_denied_handler',
}

WSGI_APPLICATION = 'ws4redis.django_runserver.application'
WEBSOCKET_URL = '/ws/'

# Internationalization
# https://docs.djangoproject.com/en/1.8/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': ('rest_framework.permissions.IsAdminUser',),
    'PAGE_SIZE': 10,
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
}

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = '/static/'

ANONYMOUS_USER_ID = None

CELERY_ACCEPT_CONTENT = ['json']
CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERYBEAT_SCHEDULER = 'djcelery.schedulers.DatabaseScheduler'

"""
##############################################################################################
#######                           USER CONFIGURATION HERE                              #######
##############################################################################################
"""

# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

WS4REDIS_SUBSCRIBER = 'cyclosible.Cyclosible.websocket.WebSocketSubscriber'
WS4REDIS_PREFIX = 'ws'
WS4REDIS_EXPIRE = 7200
WS4REDIS_CONNECTION = {
    'host': 'localhost',
    'port': 6379,
    'db': 1,
}

BROKER_URL = 'redis://localhost:6379/0'
CELERY_RESULT_BACKEND = 'redis://localhost:6379/0'
CELERY_TIMEZONE = 'Europe/Paris'

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'll=1_7y5i4e!5)@g+25ouhhz2kru6d=15twq5c)faerhj5x^@9'

# Trailing slash is needed here
PLAYBOOK_PATH = "/home/<playbook-path>/"

STORAGE_ENABLED = ['s3']

# Can be 'password', 'file' or 'hashicorp'.
# The hashicorp plugin will request the password in hashicorp's vault
# Let this variable to None to disable vault password
# Uncomment VAULT_* variables to configure the plugins
# As it's a python file, you can also use os.environ to get the
# value from your environment

VAULT_ENABLED = None
# VAULT_PASSWORD = "XXXXXXXX"
# VAULT_FILE = "/home/<vault_password_path>"
# VAULT_HASHICORP = {
#     'scheme': 'https',
#     'host': 'vault.company.com',
#     'port': 8200,
#     'token': 'xxxxxxxxxxxxxxxxx',
#     'secret_path': 'cycloid/ansible',
#     'secret_field': 'value'
# }

S3_BUCKET = "cycloid-cyclosible"
S3_ACCESS_KEY = "XXXXXXXXXXX"
S3_SECRET_KEY = "XXXXXXXXXXXXXXXXXXXXXXXXXXXX"

if 'CYCLOSIBLE_CONFIG' in os.environ:
    execfile(os.environ['CYCLOSIBLE_CONFIG'])



from cyclosible.Cyclosible.routers import main_router
from .views import (PlaybookViewSet, PlaybookRunHistoryViewSet)

main_router.register(r'playbooks', PlaybookViewSet)
main_router.register(r'playbookrunhistorys', PlaybookRunHistoryViewSet)
